#include <asm/system.h>

int set_bit(unsigned int bit, unsigned char* addr)
{
    int flags;
    int r;
    int mask;
    mask = 1 << (bit % 8);
    save_flags_cli(flags);
    r = addr[bit/8] & mask;
    addr[bit/8] |= mask;
    restore_flags(flags);
    return r;
}

int clear_bit(unsigned int bit, unsigned char* addr)
{
    int flags;
    int r;
    int mask;
    mask = 1 << (bit % 8);
    save_flags_cli(flags);
    r = addr[bit/8] & mask;
    addr[bit/8] &= ~mask;
    restore_flags(flags);
    return r;
}

int change_bit(unsigned int bit, unsigned char* addr)
{
    int flags;
    int r;
    int mask;
    mask = 1 << (bit % 8);
    save_flags_cli(flags);
    r = addr[bit/8] & mask;
    addr[bit/8] ^= mask;
    restore_flags(flags);
    return r;
}

/* lsb is bit 0 */
/* it is a lot easier in ARM to do a search_for_1 */

int find_next_zero_bit(char* addr, int maxbit, int offset)
{
    unsigned int b;
    addr += ((unsigned int)offset >> 3);
    while (1) {
        b = ~((*(signed char*)addr++) >> (offset & 7));
        if (b)
            break;
        offset = (offset & ~7) + 8;
        if (offset>=maxbit) return maxbit+1;
    }
    
    if (!(b & 0x000f)) b>>=4, offset+=4;
    if (!(b & 0x0003)) b>>=2, offset+=2;
    if (!(b & 0x0001)) offset+=1;
    return offset;
}
int find_first_zero_bit(char* addr, int maxbit)
{
    return find_next_zero_bit(addr, maxbit, 0);
}

